/*
 *  ARM64 Specific Low-Level ACPI Boot Support
 *
 *  Copyright (C) 2013-2014, Linaro Ltd.
 *	Author: Al Stone <al.stone@linaro.org>
 *	Author: Graeme Gregory <graeme.gregory@linaro.org>
 *	Author: Hanjun Guo <hanjun.guo@linaro.org>
 *	Author: Tomasz Nowicki <tomasz.nowicki@linaro.org>
 *	Author: Naresh Bhat <naresh.bhat@linaro.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 */

#define pr_fmt(fmt) "ACPI: " fmt

#include <linux/acpi.h>
#include <linux/bootmem.h>
#include <linux/cpumask.h>
#include <linux/init.h>
#include <linux/irq.h>
#include <linux/irqdomain.h>
#include <linux/memblock.h>
#include <linux/of_fdt.h>
#include <linux/smp.h>
#include <linux/console.h>
#include <linux/tty.h>

#include <asm/cputype.h>
#include <asm/cpu_ops.h>
#include <asm/smp_plat.h>

#ifdef CONFIG_ACPI_APEI
# include <linux/efi.h>
# include <asm/pgtable.h>
#endif

int acpi_noirq = 1;		/* skip ACPI IRQ initialization */
int acpi_disabled = 1;
EXPORT_SYMBOL(acpi_disabled);

int acpi_pci_disabled = 1;	/* skip ACPI PCI scan and IRQ initialization */
EXPORT_SYMBOL(acpi_pci_disabled);

static bool param_acpi_off __initdata;
static bool param_acpi_force __initdata;

static int __init parse_acpi(char *arg)
{
	if (!arg)
		return -EINVAL;

	/* "acpi=off" disables both ACPI table parsing and interpreter */
	if (strcmp(arg, "off") == 0)
		param_acpi_off = true;
	else if (strcmp(arg, "force") == 0) /* force ACPI to be enabled */
		param_acpi_force = true;
	else
		return -EINVAL;	/* Core will print when we return error */

	return 0;
}
early_param("acpi", parse_acpi);

static int __init dt_scan_depth1_nodes(unsigned long node,
				       const char *uname, int depth,
				       void *data)
{
	/*
	 * Return 1 as soon as we encounter a node at depth 1 that is
	 * not the /chosen node.
	 */
	if (depth == 1 && (strcmp(uname, "chosen") != 0))
		return 1;
	return 0;
}

u64 spcr_serial_addr;
struct acpi_device *acpi_spcr_serial_device;
EXPORT_SYMBOL(acpi_spcr_serial_device);
char *acpi_spcr_serial_options;
EXPORT_SYMBOL(acpi_spcr_serial_options);

/*
 * __acpi_map_table() will be called before page_init(), so early_ioremap()
 * or early_memremap() should be called here to for ACPI table mapping.
 */
char *__init __acpi_map_table(unsigned long phys, unsigned long size)
{
	if (!size)
		return NULL;

	return early_memremap(phys, size);
}

void __init __acpi_unmap_table(char *map, unsigned long size)
{
	if (!map || !size)
		return;

	early_memunmap(map, size);
}

bool __init acpi_psci_present(void)
{
	return acpi_gbl_FADT.arm_boot_flags & ACPI_FADT_PSCI_COMPLIANT;
}

/* Whether HVC must be used instead of SMC as the PSCI conduit */
bool __init acpi_psci_use_hvc(void)
{
	return acpi_gbl_FADT.arm_boot_flags & ACPI_FADT_PSCI_USE_HVC;
}

/*
 * acpi_fadt_sanity_check() - Check FADT presence and carry out sanity
 *			      checks on it
 *
 * Return 0 on success,  <0 on failure
 */
static int __init acpi_fadt_sanity_check(void)
{
	struct acpi_table_header *table;
	struct acpi_table_fadt *fadt;
	acpi_status status;
	acpi_size tbl_size;
	int ret = 0;

	/*
	 * FADT is required on arm64; retrieve it to check its presence
	 * and carry out revision and ACPI HW reduced compliancy tests
	 */
	status = acpi_get_table_with_size(ACPI_SIG_FADT, 0, &table, &tbl_size);
	if (ACPI_FAILURE(status)) {
		const char *msg = acpi_format_exception(status);

		pr_err("Failed to get FADT table, %s\n", msg);
		return -ENODEV;
	}

	fadt = (struct acpi_table_fadt *)table;

	/*
	 * Revision in table header is the FADT Major revision, and there
	 * is a minor revision of FADT which was introduced by ACPI 5.1,
	 * we only deal with ACPI 5.1 or newer revision to get GIC and SMP
	 * boot protocol configuration data.
	 */
	if (table->revision < 5 ||
	   (table->revision == 5 && fadt->minor_revision < 1)) {
		pr_err("Unsupported FADT revision %d.%d, should be 5.1+\n",
		       table->revision, fadt->minor_revision);
		ret = -EINVAL;
		goto out;
	}

	if (!(fadt->flags & ACPI_FADT_HW_REDUCED)) {
		pr_err("FADT not ACPI hardware reduced compliant\n");
		ret = -EINVAL;
	}

out:
	/*
	 * acpi_get_table_with_size() creates FADT table mapping that
	 * should be released after parsing and before resuming boot
	 */
	early_acpi_os_unmap_memory(table, tbl_size);
	return ret;
}

/*
 * acpi_boot_table_init() called from setup_arch(), always.
 *	1. find RSDP and get its address, and then find XSDT
 *	2. extract all tables and checksums them all
 *	3. check ACPI FADT revision
 *	4. check ACPI FADT HW reduced flag
 *
 * We can parse ACPI boot-time tables such as MADT after
 * this function is called.
 *
 * On return ACPI is enabled if either:
 *
 * - ACPI tables are initialized and sanity checks passed
 * - acpi=force was passed in the command line and ACPI was not disabled
 *   explicitly through acpi=off command line parameter
 *
 * ACPI is disabled on function return otherwise
 */
void __init acpi_boot_table_init(void)
{
	/*
	 * ACPI is disabled at this point. Leave it that way
	 * if acpi=off was seen.
	 */
	if (param_acpi_off)
		return;

	/*
	 * ACPI is disabled at this point. Enable it in order to parse
	 * the ACPI tables and carry out sanity checks
	 */
	enable_acpi();

	/*
	 * If ACPI tables are initialized and FADT sanity checks passed,
	 * leave ACPI enabled and carry on booting; otherwise disable ACPI
	 * on initialization error.
	 * If acpi=force was passed on the command line it forces ACPI
	 * to be enabled even if its initialization failed.
	 */
	if (acpi_table_init() || acpi_fadt_sanity_check()) {
		pr_err("Failed to init ACPI tables\n");
		if (!param_acpi_force)
			disable_acpi();
	}
}

#ifdef CONFIG_ACPI_APEI
pgprot_t arch_apei_get_mem_attribute(phys_addr_t addr)
{
	/*
	 * According to "Table 8 Map: EFI memory types to AArch64 memory
	 * types" of UEFI 2.5 section 2.3.6.1, each EFI memory type is
	 * mapped to a corresponding MAIR attribute encoding.
	 * The EFI memory attribute advises all possible capabilities
	 * of a memory region. We use the most efficient capability.
	 */

	u64 attr;

	attr = efi_mem_attributes(addr);
	if (attr & EFI_MEMORY_WB)
		return PAGE_KERNEL;
	if (attr & EFI_MEMORY_WT)
		return __pgprot(PROT_NORMAL_WT);
	if (attr & EFI_MEMORY_WC)
		return __pgprot(PROT_NORMAL_NC);
	return __pgprot(PROT_DEVICE_nGnRnE);
}
#endif

bool acpi_spcr_console_check(struct acpi_device *adev, char *name, int index)
{
	if (!adev || adev != acpi_spcr_serial_device || console_set_on_cmdline)
		return false;

	pr_info("adding preferred console [%s]\n", name);

	return !add_preferred_console(name, index,
				      kstrdup(acpi_spcr_serial_options,
						GFP_KERNEL));
}

/*
 * Parse the SPCR table. If we are not working with version 2 or
 * higher, bail.
 * Otherwise, pull out the baud rate and address to the console device.
 */
static int __init acpi_parse_spcr(struct acpi_table_header *table)
{
	struct acpi_table_spcr *spcr = (struct acpi_table_spcr *)table;

	if (table->revision < 2)
		return -EOPNOTSUPP;

	/* Handle possible alignment issues */
	memcpy(&spcr_serial_addr,
	       &spcr->serial_port.address, sizeof(spcr_serial_addr));

	/*
	 * The baud rate the BIOS used for redirection. Valid values are....
	 *	3 = 9600
	 *	4 = 19200
	 *	6 = 57600
	 *	7 = 115200
	 *	0-2, 5, 8 - 255 = reserved
	*/
	switch (spcr->baud_rate) {
	case 3:
		acpi_spcr_serial_options = "9600";
		break;
	case 4:
		acpi_spcr_serial_options = "19200";
		break;
	case 6:
		acpi_spcr_serial_options = "57600";
		break;
	case 7:
		acpi_spcr_serial_options = "115200";
		break;
	default:
		acpi_spcr_serial_options = "";
		break;
	}

	pr_info("SPCR serial device: 0x%llx (options: %s)\n",
	       spcr_serial_addr, acpi_spcr_serial_options);

	return 0;
}

/*
 * Parse an ACPI "Device" to determine if it represents the
 * data found in the SPCR table. If the associated Device has
 * and Address entry, and, that Address matches the one found
 * in our SPCR table, it's the entry we are interested in.
 *
 */
static acpi_status acpi_spcr_device_scan(acpi_handle handle,
					 u32 level, void *context, void **retv)
{
	struct acpi_buffer name_buffer = { ACPI_ALLOCATE_BUFFER, NULL };
	acpi_status status = AE_OK;
	struct acpi_device *adev;
	struct list_head resource_list;
	struct resource_entry *r;

	adev = acpi_bus_get_acpi_device(handle);
	if (!adev) {
		pr_err("Err locating SPCR device from ACPI handle\n");
		return AE_OK; /* skip this one */
	}

	status = acpi_get_name(handle, ACPI_FULL_PATHNAME, &name_buffer);
	if (ACPI_FAILURE(status))
		goto ret;

	/*
	 * Look for spcr_serial_addr in resource list.
	 */
	INIT_LIST_HEAD(&resource_list);
	if (acpi_dev_get_resources(adev, &resource_list, NULL, NULL) <= 0)
		goto ret;

	list_for_each_entry(r, &resource_list, node) {
		if (resource_type(r->res) == IORESOURCE_MEM &&
		    r->res->start == spcr_serial_addr) {
			acpi_spcr_serial_device = adev;

			pr_info("SPCR serial console: %s (0x%llx)\n",
				(char *)(name_buffer.pointer),
				spcr_serial_addr);
			break;
		}
	}
	acpi_dev_free_resource_list(&resource_list);

ret:
	acpi_os_free(name_buffer.pointer);
	return status;
}

int __init acpi_setup_spcr(void)
{
	if (0 != acpi_table_parse(ACPI_SIG_SPCR, acpi_parse_spcr)) {
		pr_warn("SPCR table not found - auto console disabled\n");
		return -ENODEV;
	}

	acpi_walk_namespace(ACPI_TYPE_DEVICE, ACPI_ROOT_OBJECT,
			    ACPI_UINT32_MAX, acpi_spcr_device_scan,
			    NULL, NULL, NULL);

	return 0;
}

static int __is_acpi_ram(u64 start, u64 end, void *arg)
{
	return 1;
}

int page_is_acpi_ram(unsigned long pfn)
{
	u64 start, end;

	start = (u64) pfn << PAGE_SHIFT;
	end = start + PAGE_SIZE - 1;

	return walk_iomem_res("ACPI RAM", IORESOURCE_MEM | IORESOURCE_BUSY,
			      start, end, NULL, __is_acpi_ram) == 1;
}
