/*
 * Extensible Firmware Interface
 *
 * Based on Extensible Firmware Interface Specification version 2.4
 *
 * Copyright (C) 2013, 2014 Linaro Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/dmi.h>
#include <linux/efi.h>
#include <linux/init.h>

#include <asm/efi.h>
#include <linux/bootmem.h>

static int __init is_normal_ram(efi_memory_desc_t *md)
{
	if (md->attribute & EFI_MEMORY_WB)
		return 1;
	return 0;
}


int __init efi_create_mapping(struct mm_struct *mm, efi_memory_desc_t *md)
{
	pteval_t prot_val;

	/*
	 * Only regions of type EFI_RUNTIME_SERVICES_CODE need to be
	 * executable, everything else can be mapped with the XN bits
	 * set.
	 */
	if ((md->attribute & EFI_MEMORY_WB) == 0)
		prot_val = PROT_DEVICE_nGnRE;
	else if (md->type == EFI_RUNTIME_SERVICES_CODE ||
		 !PAGE_ALIGNED(md->phys_addr))
		prot_val = pgprot_val(PAGE_KERNEL_EXEC);
	else
		prot_val = pgprot_val(PAGE_KERNEL);

	create_pgd_mapping(mm, md->phys_addr, md->virt_addr,
			   md->num_pages << EFI_PAGE_SHIFT,
			   __pgprot(prot_val | PTE_NG));
	return 0;
}

#ifdef CONFIG_ACPI
static void __init __insert_resource(struct resource *res)
{
	struct resource *r, *conflict;

	r = alloc_bootmem_low(sizeof(*r));
	*r = *res;

	do {
		conflict = insert_resource_conflict(&iomem_resource, r);
		if (conflict) {
			struct resource *new;

			/*
			 * Split into region outside conflict and region
			 * inside conflict.
			 */
			new  = alloc_bootmem_low(sizeof(*new));
			*new = *r;

			if (r->start >= conflict->start) {
				r->start = conflict->end + 1;
				new->end = conflict->end;
			} else {
				r->end = conflict->start - 1;
				new->start = conflict->start;
			}
			insert_resource(&iomem_resource, new);
		}
	} while (conflict);
}

/*
 * Create /proc/iomem resources for any ACPI regions in RAM.
 */
void __init efi_request_acpi_resources(void)
{
	struct resource res;
	efi_memory_desc_t *md;
	u64 start, end, npages;
	unsigned long mapsize = memmap.map_end - memmap.map;
	void *map;

	map = early_memremap((resource_size_t)memmap.phys_map, mapsize);
	if (map == NULL)
		return;
	memmap.map = map;
	memmap.map_end = map + mapsize;

	memset(&res, 0, sizeof(res));
	res.name  = "ACPI RAM";
	res.flags = IORESOURCE_MEM | IORESOURCE_BUSY;

	for_each_efi_memory_desc(&memmap, md) {
		if (!is_normal_ram(md) ||
		    !(md->type == EFI_ACPI_RECLAIM_MEMORY ||
		      md->type == EFI_ACPI_MEMORY_NVS))
			continue;

		start = md->phys_addr;
		npages = md->num_pages;
		memrange_efi_to_native(&start, &npages);
		end = start + (npages << PAGE_SHIFT) - 1;

		if (res.end == 0) {
			res.start = start;
			res.end = end;
			continue;
		}

		if (start >= res.start && (start - 1) <= res.end) {
			/* overlaps (or adjacent to) end of old region */
			if (end > res.end)
				res.end = end;
		} else if (end >= (res.start - 1) && end <= res.end) {
			/* overlaps (or adjacent to) start of old region */
			if (start  < res.start)
				res.start = start;
		} else {
			__insert_resource(&res);
			res.start = start;
			res.end = end;
		}
	}
	if (res.end)
		__insert_resource(&res);

	early_memunmap(memmap.map, mapsize);
}
#endif

static int __init arm64_dmi_init(void)
{
	/*
	 * On arm64, DMI depends on UEFI, and dmi_scan_machine() needs to
	 * be called early because dmi_id_init(), which is an arch_initcall
	 * itself, depends on dmi_scan_machine() having been called already.
	 */
	dmi_scan_machine();
	if (dmi_available)
		dmi_set_dump_stack_arch_desc();
	return 0;
}
core_initcall(arm64_dmi_init);

/*
 * UpdateCapsule() depends on the system being shutdown via
 * ResetSystem().
 */
bool efi_poweroff_required(void)
{
	return efi_enabled(EFI_RUNTIME_SERVICES);
}
